#!/usr/bin/env python
# coding: utf-8


# ESA (C) 2000-2021
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# om_background_plots.py

import pysas.pyutils.pyutils as pyutils
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from matplotlib.ticker import PercentFormatter
from scipy.stats import poisson
import re
import glob
import sys
import warnings
import pickle
import pysas.pysasplot_utils.pysasplot_utils as sasplt
import os


def read_datafile(path_to_data):
    """
    Opens the input data file and returns a NumPy array.

    Args:
        path_to_data: the path to the data file (expected a txt).

    Output:
        data: a NumPy array containing the data.
    """

    data = np.loadtxt(path_to_data)
    return data


def om_background_fit(data, plot_title, filename, add_text = '', nbins = 10, scale = ''):
    """
    Saves and plots (if possible) the histogram for the pixel counts of the
    opitcal monitor background. Will also add Poisson and Gaussian fits
    alongside the histogram and display the needed parameters from the fits.

    Args:
        data: the data as an array, list or other comprehensive types for
    matplotlib.
        filename: the name used to save the file.
        add_text: adds text to the figure. If left empty, the plot will take
    the non-used space.
        bins: int or tuple, in case that the standard calculated by plt.hist
    are not satisfying (10 by default).
        scale: the scale to apply. None by default. Check the official plt docs
    to get a better insight of the available scales.
    """

    if isinstance(data, str):
        data = read_datafile(data)

    if nbins == '':
        nbins = np.arange(max(data)) - 0.5
    
    data = data.flatten()
    data = return_filtered_data(data)
    fig = plt.figure(figsize = (20, 17), facecolor = 'white')
    ax = fig.add_axes([0.3,0.3,0.6,0.3])
    
    entries, bin_edges, patches = ax.hist(data, weights=np.ones(len(data)) / len(data), bins = nbins, color = 'forestgreen', fill = True, alpha = 0.2, density = True, label='Data')

    # calculate bin centres
    bin_middles = 0.5 * (bin_edges[1:] + bin_edges[:-1])


    def fit_function(k, lamb):
        """
        Poisson function, parameter lamb is the fit parameter

        Args:
            k: the fitting data.
            lamb: the fit parameter.

        Output:
            the Poissonian probability mass function for the data.
        """

        poisson_fit_data = poisson.pmf(k, lamb)
        return poisson_fit_data


    def gauss(x, *p):
        """
        Returns a Gauss-like curve fit with the input data and the parameters
        passed as input.

        Args:
            x: the fitting data.
            p: encapsulated parameters.

        Output:
            a Gaussian curve fit (numpy.ndarray class).
        """

        A, mu, sigma = p
        return A * np.exp(-(x - mu)**2 / (2 * sigma ** 2))


    # fit with curve_fit
    mean_estimator = sum(data) / len(data)
    int_bins = [int(x) for x in bin_middles] # best fit for a Poissonian distribution
    
    parameters, cov_matrix = curve_fit(fit_function, int_bins, entries, mean_estimator)
    
    gaussian_sigma_estimator = np.sqrt(sum((np.ones(len(data)) * mean_estimator - data)**2) / (len(data) - 1))
    
    p0 = (1, mean_estimator, gaussian_sigma_estimator) # first Gaussian estimator
    parameters_g, cov_matrix_g = curve_fit(gauss, bin_middles, entries, p0)

    # plot poisson-deviation with fitted parameter
    min_x_plot = int(bin_edges[0])
    if min_x_plot < 0:
        min_x_plot = 0
    x_plot = np.arange(min_x_plot, 1 + int(bin_edges[-1]))

    plt.plot(
        x_plot,
        fit_function(x_plot, *parameters),
        marker = '.', linestyle = '-',
        label = 'Poissonian fit.',
    )
    plt.plot(
        x_plot,
        gauss(x_plot, *parameters_g),
        marker = '.', linestyle = '-',
        label = 'Gaussian fit.',
    )

    plt.legend()

    lambda_err = round(np.sqrt(cov_matrix[0][0]), 2)
    gaussian_errors = np.sqrt(np.diag(cov_matrix_g))

    lamb = round(parameters[0], 2)
    median_value_p = (round(lamb - np.log(2), 2), round(lamb + 1/3, 2))
    std_dev_p = round(np.sqrt(parameters[0]), 2)
    std_dev_p_err = round(lambda_err / (2 * np.sqrt(lamb)), 2)
    std_dev_g = round(parameters_g[2], 2)
    mu_g = round(parameters_g[1], 2)
    mu_g_err = round(gaussian_errors[1], 2)
    std_dev_g_err = round(gaussian_errors[2], 2)

    stats_text = r'Mean = {}'.format(round(mean_estimator,3)) + '\n' + r'$\sigma = {}$'.format(round(gaussian_sigma_estimator, 2))
    text_gauss = r'$\mu = {} \pm {}$'.format(mu_g, mu_g_err) + '\n' + r'$\sigma = {} \pm {}$'.format(std_dev_g, std_dev_g_err)
    text_poisson = r'$\lambda = {} \pm {}$'.format(lamb, lambda_err) + '\n' + 'Poisson median interval:\n' + r'${0} \leq {1} < {2}$'.format(median_value_p[0], poisson.median(mu = lamb), median_value_p[1]) + '\n' + r'$\sigma = {} \pm {}$'.format(std_dev_p, std_dev_p_err)
    add_text = 'Data stats:\n' + stats_text.strip() + '\n\n' + 'Gaussian fit:\n' + text_gauss.strip() + '\n\n' + 'Poissonian fit:\n' + text_poisson.strip()
    plt.gcf().text(0.02, 0.3, add_text, fontsize=14)
    plt.gca().yaxis.set_major_formatter(PercentFormatter(1))
    plt.subplots_adjust(top = 0.2)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    plt.xlabel('Pixel counts')
    plt.ylabel('Frequency (%) (N = {})'.format(len(data)))
    plt.title(plot_title)

    file_name = re.sub(r'\W+', '', filename)
        # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        fig.savefig("{}_1.pdf".format(file_name), bbox_inches='tight')
        #pickle.dump(fig, open('{}_1.fig'.format(file_name), 'wb'))
    else:
        fig.savefig("{}_1.pdf".format(file_name), bbox_inches='tight')
        #pickle.dump(fig, open('{}_1.fig'.format(file_name), 'wb'))


def ombackground_hists(y_data, ndata, plot_title, filename, x_axis, nbins = 50, add_text = '', y_labels = ('1', '2', '3', '4',), scale = ''):
    """
    Plotting device for ombackground histograms.

    Args:
        y_data: set of arrays, containing four different histogram data. Also support txt files.
        ndata: the number of data points per method, specified in txt files.
        plot_title: the title for the plot. By default, om_back_hists.
        filename: the name used to save the file.
        x_axis: the text to add to the x axis.
        nbins: the number of bins for the histogram. Empty by default (derived
    from np.histogram after stacking the total data assuming 10 bins).
        add_text: additional text to add.
        y_labels: a set of labels with the same size as the y_data set of arrays.
        scale: empty by default, the scaling for the y axis.
    """

    fig = plt.figure(figsize = (20, 17), facecolor = 'white')
    ax = fig.add_axes([0.3,0.3,0.6,0.3])
    
    # If y_data is a set of filenames then readin the data
    if sasplt.is_iterable(y_data):
        loaded_data = []
        for i in y_data:
            if isinstance(i, str):
                loaded_data.append(read_datafile(i))
            else:
                loaded_data.append(i)
        y_data = loaded_data
        del(loaded_data)
        #print('A: ',y_data.size,y_data.ndim)
    else:
        if isinstance(y_data, str):
            y_data = read_datafile(y_data)
            #print('B: ',y_data.size,y_data.ndim)

    #print('C2: ',y_data.size,y_data.ndim)
    if ndata is None:
        pass
    else:
        ndata = np.loadtxt(ndata)
        y_data = y_data.reshape(int(ndata[0]), len(ndata)).T

    #bins = np.histogram(np.hstack(y_data), bins=nbins)[1]
    bins = nbins
    
    for col in range(0, len(y_data)):
        try:
            y_data_col = y_data[col]
            y_data_col = return_filtered_data(y_data_col)
            if col == 0:
                _, bins, _ = ax.hist(y_data_col, label = str(y_labels[col]), histtype='step', bins = bins,fill = False, alpha = 1)
            else:
                ax.hist(y_data_col, label = str(y_labels[col]), histtype='step', bins = bins, fill = False, alpha = 1)
        except IndexError:
            warnings.warn('At least one of the histograms/labels was not complete.')
    
    #ax.legend(bbox_to_anchor = (1.05, 1), loc = 'upper left', borderaxespad=0.)
    ax.legend(loc = 'best', frameon=False, fontsize='medium')
    
    plt.gcf().text(0.02, 0.3, add_text, fontsize=14)
    plt.subplots_adjust(top = 0.2)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    plt.xlabel(x_axis)
    plt.ylabel('Number')
    plt.title(plot_title)

    file_name = re.sub(r'\W+', '', filename)
    # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        #fig.savefig("{}.pdf".format(file_name), bbox_inches='tight')
        if os.path.isfile('{}_1.fig'.format(file_name)):
            file_name = file_name + '1'
            #pickle.dump(fig, open('{}_1.fig'.format(file_name), 'wb'))
            fig.savefig("{}.pdf".format(file_name), bbox_inches='tight')
        else:
            fig.savefig("{}.pdf".format(file_name), bbox_inches='tight')
            #pickle.dump(fig, open('{}_1.fig'.format(file_name), 'wb'))
    else:
        if os.path.isfile('{}_1.fig'.format(file_name)):
            file_name = file_name + '1'
            fig.savefig("{}.pdf".format(file_name), bbox_inches='tight')
            #pickle.dump(fig, open('{}_1.fig'.format(file_name), 'wb'))
        else:
            #with open('{}_1.fig'.format(file_name), 'wb') as f:
            #    pickle.dump(fig, f)
            fig.savefig("{}.pdf".format(file_name), bbox_inches='tight')


def pdf_merge(output_file):
    """
    Merges all the PDF files found in the current working directory.

    Args:
        output_file: the name of the final pdf.
    """

    pdf_list = glob.glob(output_file + '*.pdf')
    sasplt.merge_pdf(pdf_list, output_file + '.pdf')
    temp_list = glob.glob('temp*')

    if len(pdf_list) != 0:
        for i in pdf_list:
            try:
                os.remove(i)
            except FileNotFoundError:
                pass
    
    if len(temp_list) != 0:
        for j in temp_list:
            try:
                #os.remove(j)
                pass
            except FileNotFoundError:
                pass


def return_filtered_data(data):
    """
    Filters the data to remove the extreme outliers corresponding to the 95 and 5 percnetiles.

    Args:
        data: the data to be filtered.

    Output:
        data: the filtered data.
    """

    #while abs(np.max(data)) > 50 * abs(np.mean(data)):
    data = data[data < np.percentile(data, 95)]
    
    #while abs(np.min(data)) < 10 * abs(np.mean(data)):
    data = data[data > np.percentile(data, 5)]

    return data


def curate_caption(text_file):
    """
    Filters the info from the texts files produced by ombackground.

    Args:
        text_file: the path to the text file.

    Ouput:
        final_txt: the text filtered and separated.
    """

    with open(text_file) as txtf:
        lines = txtf.readlines()
        txt = lines[0]

    # The \s substitution is a syntax error - now commented out
    #txt = re.sub("\s\s+" , " ", txt)
    txt = txt.replace('+/-', u"\u00B1")

    mean_pos = txt.find('Mean')
    txt2 = txt[:mean_pos] + '\n' + txt[mean_pos:]

    med_pos = txt2.find('Median')
    txt3 = txt2[:med_pos] + '\n' + txt2[med_pos:]

    mode_pos = txt3.find('Mode')
    txt4 = txt3[:mode_pos] + '\n' + txt3[mode_pos:]

    label_pos = txt4.find('xLabel')
    txt5 = txt4[:label_pos] + '\n' + txt4[label_pos:]

    title_pos = txt5.find('title')
    final_txt = txt5[:title_pos] + '\n' + txt5[title_pos:]

    txt_splitted = final_txt.split('\n')
    captions = txt_splitted[0:4]
    xlabel = txt_splitted[-3]
    title = txt_splitted[-2]

    xlabel = xlabel.replace('xLabel:', '')
    title = title.replace('title', '')
    
    return captions, xlabel, title


################################################
################### Run code ###################
################################################


if __name__ == '__main__':
    nbins = int(sys.argv[1])
    to_fit_data = glob.glob('fitted_data*')
    # Get the filename to use.
    try:
        with open('plotname.txt') as pltname:
            filename = pltname.readlines()[0].strip()
            filename = filename.split('.')[0] # removes extension (if any)
    except FileNotFoundError:
            print('Could not locate plotname.txt. Using default filename: ombackground.')
            filename = 'ombackground'
    try:
        os.remove('plotname.txt')
    except FileNotFoundError:
        pass


    # First histogram and poissonian/gaussian fit
    if len(to_fit_data) > 0:
        for i in to_fit_data:
            with open('fitted_info.txt') as info_t:
                plot_title = info_t.readlines()[0].strip()
            om_background_fit(i, plot_title, filename, add_text = '', nbins = nbins, scale = '')
    else:
        warnings.warn('Could not find suitable data for fitting histogram.')


    # Second function: several histograms.
    ndata_files = glob.glob('ndataM*')
    ndata_files.sort()
    backg_files = glob.glob('backgM*')
    backg_files.sort()
    caption_files = glob.glob('ncapti*')
    caption_files.sort()
    
    for i in range(0, len(backg_files)):
        y_data = backg_files[i]
        captions, xlabel, plot_title = curate_caption(caption_files[i])
        ndata_f = ndata_files[i]
        ombackground_hists(y_data, ndata_f, plot_title, filename + '{}'.format(i), xlabel, nbins = nbins, add_text = '', y_labels = captions, scale = '')
    

    # Merging all created pdfs.
    # If an older version exists, it will be removed first.
    if os.path.isfile(filename + '.pdf'):
        os.remove(filename + '.pdf')

    pdf_merge(filename)


    # Cleaning up...
    os.remove('fitted_info.txt')
    os.remove('fitted_data.txt')
    for i in ndata_files:
        os.remove(i)

    for i in caption_files:
        os.remove(i)

    for i in backg_files:
        os.remove(i)
